
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from qutree.ttn.grid import *
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import imageio
import os

def plot_xyz(xyz, f, ranges = None):
    # Create a 3D scatter plot with colors based on f_values
    xyz = xyz.grid
    fig = go.Figure(data=[go.Scatter3d(
        x=xyz[:, 0],
        y=xyz[:, 1],
        z=xyz[:, 2],
        mode='markers',
        marker=dict(
            size=5,
            color=f,  # Use f_values as colors
            colorscale='Viridis',  # Choose a colorscale
            colorbar=dict(title='Function Value')  # Add colorbar with a title
        )
    )])

    # Set plot layout with fixed axis range
    if ranges is None:
        fig.update_layout(
            scene=dict(
                xaxis=dict(title='X Axis', autorange=True),  # Fix x-axis range to [0, 1]
                yaxis=dict(title='Y Axis', autorange=True),  # Fix y-axis range to [0, 1]
                zaxis=dict(title='Z Axis', autorange=True),   # Fix z-axis range to [0, 1]
                aspectmode='cube',  # Ensure that aspect ratio is maintained
            ),
            margin=dict(l=0, r=0, b=0, t=0)
        )
    else:
        fig.update_layout(
            scene=dict(
                xaxis=dict(title='X Axis', range=ranges[0], autorange=False),  # Fix x-axis range to [0, 1]
                yaxis=dict(title='Y Axis', range=ranges[1], autorange=False),  # Fix y-axis range to [0, 1]
                zaxis=dict(title='Z Axis', range=ranges[2], autorange=False),   # Fix z-axis range to [0, 1]
                aspectmode='cube',  # Ensure that aspect ratio is maintained
            ),
            margin=dict(l=0, r=0, b=0, t=0)
        )

    # Show the plot
    return fig

def plot_tt_diagram(tn, draw_ranks = True):
    nleaves = len(up_leaves(tn))
    pos0 = [np.array([-(-i - 1), 0.1]) for i in range(-nleaves, 0)]
    pos1 = [np.array([-i, 0]) for i in range(nleaves)]
    pos12 = pos0 + pos1
    pos = {i: pos12[i] for i in range(-nleaves, nleaves)}
#    plt.draw() 
    plt.gca().set_aspect(15)  # 'equal' ensures that one unit in x is equal to one unit in y
    fig = nx.draw(tn, pos = pos, with_labels=False, node_size = 250, font_size = 8)
    if draw_ranks:
        ranks = nx.get_edge_attributes(tn, 'r')
        nx.draw_networkx_edge_labels(tn, pos, edge_labels=ranks, font_size=14)
#    plt.subplots_adjust(left=0.1, right=0.9, bottom=10.1, top=20.9)
    return fig

def plot_tn_xyz(tn, fun, q_to_x = None):
    # collect grids
    gs = nx.get_node_attributes(tn, 'grid')
    gs = list(gs.values()) # dict to list
    grid = direct_sum(gs)
    if not q_to_x is None:
        grid = grid.transform(q_to_x)
    fig = plot_xyz(grid, grid.evaluate(fun))
    fig.show()

def tn_to_df(tn, fun):
    # collect grids
    gs = nx.get_node_attributes(tn, 'grid')
    gs = list(gs.values()) # dict to list
    xyz, f, node = [], [], []
    for id, grid in enumerate(gs):
        for i, point in enumerate(grid.grid):
            xyz.append(point)
            node.append(id)
    return pd.DataFrame({'xyz': xyz, 'node': node})

def plot_tree(G, draw_ranks = True):
    G = add_layer_index(G)
    nleaves = len(up_leaves(G))
    grid = np.linspace(0, 1, nleaves)
    pos = {i : (0, 0) for i in sorted(G.nodes)}
    for node in sorted(G.nodes):
        layer = G.nodes[node]["layer"]
        y = -layer
        x = 0.
        if node < 0:
            id = -node - 1
            x = grid[id]
        else:
            cs = children(G, node)
            children_pos = np.array([pos[c][0] for c in cs])
            x = np.mean(children_pos, axis=0)
        pos[node] = (x, y)
    
    nx.draw(G, pos = pos, with_labels=False, node_size = 500)
    if draw_ranks:
        ranks = nx.get_edge_attributes(G, 'r')
        nx.draw_networkx_edge_labels(G, pos, edge_labels=ranks, font_size = 18)
    plt.draw()

    
def tngrid_to_df(G, O):
    # Get node attributes as a dictionary
    node_attributes = nx.get_node_attributes(G, 'grid')
    node_attributes = {k: v.grid for k, v in node_attributes.items() if v is not None}

    # Convert dictionary to DataFrame
    df = pd.DataFrame(list(node_attributes.items()), columns=['node', 'grid'])

    # Create a new DataFrame with the exploded values and reset the index
    df = df.explode('grid').reset_index()

    df['f'] = df['grid'].apply(lambda x: O.Err(x))

    # Remove the 'index' column
    df = df.drop(columns='index')

    # Convert the 'grid' column into separate columns
    L = len(df['grid'].iloc[0])
    df[['x{}'.format(i+1) for i in range(L)]] = df['grid'].apply(pd.Series)

    # Drop the original 'grid' column
    df.drop(columns='grid', inplace=True)
    return df

def concat_pandas(dfs):
    for t, df in enumerate(dfs):
        df['time'] = t
    df = pd.concat(dfs).reset_index()
    df["size"] = 1
    return df

def grid_animation(df, color = 'f'):
    fig = px.scatter_3d(df, x="x1", y="x2", z="x3", animation_frame="time", animation_group="node",
               size="size", color=color, hover_name="time",
               size_max=15,
                width=1000, height=800)
    fig.layout.updatemenus[0].buttons[0].args[1]['frame']['duration'] = 100
    fig.layout.updatemenus[0].buttons[0].args[1]['transition']['duration'] = 20
    fig.update_layout(scene=dict(
        xaxis_title="x",
        yaxis_title="y",
        zaxis_title="z",
        xaxis=dict(showticklabels=False),
        yaxis=dict(showticklabels=False),
        zaxis=dict(showticklabels=False),
    ))

    camera = dict(
        eye=dict(x=1.5, y=-1.0, z=1.2),  # Set the position of the camera
        center=dict(x=0, y=0, z=0),       # Set the point the camera is looking at
        up=dict(x=0, y=0, z=1)            # Set the upward direction of the camera
    )
    fig.update_layout(scene_camera=camera)
    return fig

def grid_animation_to_gif(df, color='f', gif_filename='animation.gif', frames_folder = '.frames'):
    os.makedirs(frames_folder, exist_ok=True)
    unique_times = df['time'].unique()
    unique_times = unique_times[2:35]
    for time in unique_times:
        sub_df = df[df['time'] == time]

        # Generate the plot
        fig = px.scatter_3d(sub_df, x="x1", y="x2", z="x3", color=color, size_max=10, width=1000, height=800)
        fig.update_layout(scene=dict(xaxis_title="x", yaxis_title="y", zaxis_title="z",
                                     xaxis=dict(showticklabels=False),
                                     yaxis=dict(showticklabels=False),
                                     zaxis=dict(showticklabels=False)))
        camera = dict(
            eye=dict(x=1.5, y=-1.0, z=1.2),  # Set the position of the camera
            center=dict(x=0, y=0, z=0),       # Set the point the camera is looking at
            up=dict(x=0, y=0, z=1)            # Set the upward direction of the camera
        )
        
        # Adjust transparency settings if needed
        fig.update_traces(marker=dict(opacity=0.75, size = 6))  # Adjust opacity here
        fig.update_layout(scene_camera=camera)
        fig.write_image(os.path.join(frames_folder, f'frame_{time:04d}.png'), format = 'png')

    # Generate GIF
    images = [imageio.imread(os.path.join(frames_folder, f'frame_{i:04d}.png')) for i in unique_times]
    imageio.mimsave(gif_filename, images, fps=10)  # Adjust fps as needed

    # Cleanup images
    for filename in os.listdir(frames_folder):
        os.remove(os.path.join(frames_folder, filename))
    os.rmdir(frames_folder)

    return gif_filename


import numpy as np
from .network import *

class quTensor(np.ndarray):
    """
    Decorated tensor that keeps track of corresponding edges
    edges: those correspond to the tensor legs
    flattened_to: None or edge that the current tensor is flattened to.
    expanded_shape: shape if not permuted and flattened

    Note: edges & expanded_shape is not permuted. Only 
    """
    def __new__(cls, array, edges, flattened_to = None):
        if (len(edges) != len(array.shape)):
            raise ValueError("Number of edges does not match the shape of the tensor.")
        obj = np.asarray(array).view(cls)
        obj.edges = [tuple(sorted(edge)) for edge in edges] # edge = (small, large)
        return obj

    def flatten(self, edge):
        edge = tuple(sorted(edge))
        p = back_permutation(self.edges, edge)
        A = self.transpose(p)
        s = [self.shape[i] for i in p]
        edges_p = [self.edges[i] for i in p]
        edges = [edges_p[:-1], edge]
        return quTensor(A.reshape((-1, s[-1])), edges)
    
    def transpose(self, axes = None):
        if axes is None:
            axes = list(range(len(self.shape)))
            axes = axes[::-1]
        edges = [self.edges[i] for i in axes]
        A = super().transpose(axes)
        return quTensor(A, edges)

def tensordot(A, B, edge):
    e = tuple(sorted(edge))
    iA = A.edges.index(e)
    iB = B.edges.index(e)
    edges_a = A.edges.copy()
    edges_a.remove(e)
    edges_b = B.edges.copy()
    edges_b.remove(e)
#    print(edges_a, edges_b)
    edges_c = edges_a + edges_b
    for i, ex in enumerate(edges_c):
        if (ex[1] == e[0]):
#            print('swapping', ex, ', new edge: ', (ex[0], e[1]))
            edges_c[i] = (ex[0], e[1])
    edges_c = [tuple(sorted(edge)) for edge in edges_c]
#    print(iA, iB)

    return quTensor(np.tensordot(A, B, axes = (iA, iB)), edges_c)



import networkx as nx
import numpy as np

"""
Network Utilities
"""

def pre_edges(G, edge, remove_flipped = False):
    pre = list(G.in_edges(edge[0]))
    if remove_flipped:
        pre.remove(flip(edge))
    return pre

def is_leaf(edge, G):
    return G.in_degree(edge[0]) <= 1 or G.in_degree(edge[1]) <= 1
#    return G.in_degree(edge[0]) == 0 or G.in_degree(edge[1]) == 0 # works only if leaves are added as (-i - 1, i) but not reverse

def is_leaf_node(node, G):
    return G.in_degree(node) <= 1 or G.in_degree(node) <= 1

def up_edge(edge, G):
    d0 = nx.shortest_path_length(G, source=edge[0], target=root(G))
    d1 = nx.shortest_path_length(G, source=edge[1], target=root(G))
    return d0 > d1

def flip(edge):
    return (edge[1], edge[0])

def back_permutation(edges, edge):
    el = edges.index(edge)
    p = list(range(len(edges)))
    p.remove(el)
    p.append(el)
    return p

def flatten_back(A, shape):
    return A.reshape((-1, shape[-1]))

def collect(G, edges, key):
    """
    Graph objects from edges that correspond to 'key' as a list
    """
    items = []
    for e in edges:
        items.append(G.edges[e][key])
    return items

def up_edges_by_distance_to_root(G, root):
    distance = nx.shortest_path_length(G, source=root)

    up_edges = [edge for edge in G.edges() if distance[edge[0]] > distance[edge[1]]]
    down_edges = [edge for edge in G.edges() if distance[edge[0]] < distance[edge[1]]]

    sorted_up_edges = sorted(up_edges, key=lambda edge: -distance[edge[1]])
    sorted_down_edges = sorted(down_edges, key=lambda edge: distance[edge[1]])
    return sorted_up_edges, sorted_down_edges

def sweep(G, include_leaves = True):
    up, down = up_edges_by_distance_to_root(G, root(G))
    sw = up + down
    if not include_leaves:
        sw = [edge for edge in sw if not is_leaf(edge, G)]
    return sw

def up_sweep(G, include_leaves = True):
    # @todo: add unit test
    up, _ = up_edges_by_distance_to_root(G, root(G))
    sw = up
    if not include_leaves:
        sw = [edge for edge in sw if not is_leaf(edge, G)]
    return sw

#def sweep(G, include_leaves = True):
#    up = sorted(G.edges, key = lambda x: x[0])
#    up = [edge for edge in up if up_edge(edge)]
#    down = reversed(up)
#    down = [edge for edge in down if not is_leaf(edge, G)]
#    down = [flip(edge) for edge in down]
#    res = up + down
#    if not include_leaves:
#        res = [edge for edge in res if not is_leaf(edge, G)]
#    return res

def rsweep(G):
    return reversed(sweep(G))

def add_leaves(G, f):
    for i in range(f):
        edge = (i, -i - 1)
        G.add_edge(i, -i - 1)
        G.add_edge(-i - 1, i)
        G.edges[edge]['coordinate'] = i
        G.edges[flip(edge)]['coordinate'] = i

def root(G):
    """
    Return the root of a tree
    """
    # @todo: remove dependency of node indexing. Maybe at least use '0' as root
    return max(G.nodes)

def up_leaves(G):
    """
    Return the leaves of a tree
    """
    return [edge for edge in G.edges if is_leaf(edge, G) and up_edge(edge, G)]

def children(G, node):
    in_edges = G.in_edges(node)
    return [e[0] for e in in_edges if e[0] < node]

def _star_sweep(G, node, parent = None, sweep = []):
    childs = children(G, node)
    for child in childs:
        if child < 0:
            continue
        sweep.append((node, child))
        _star_sweep(G, child, node, sweep)
    if parent is not None:
        sweep.append((node, parent))

def star_sweep(G, exclude_leafs = False):
    rt = root(G)
    sweep = []
    _star_sweep(G, rt, None, sweep)
    if exclude_leafs:
        sweep = [edge for edge in sweep if not is_leaf(edge, G)]
    return sweep

def remove_edge(G, edge):
    pre = pre_edges(G, edge, remove_flipped=True)
    v = edge[0]
    w = edge[1]
    # redirect edges
    for e in pre:
        attr = G.get_edge_data(*e)
        G.add_edge(e[0], w, **attr)
        G.remove_edge(*e)
    G.remove_edge(*edge)
    G.remove_edge(*flip(edge))
    G.remove_node(v)
    return G

"""
Tensor Networks
"""

def add_layer_index(G, root = None):
    if root is None:
        root = max(G.nodes)
    for node in G.nodes:
        layer = nx.shortest_path_length(G, node, root)
        G.nodes[node]['layer'] = layer
    return G

def tensor_train_graph(f, r = 2, primitive_grid: int | list[int] = 8):
    """
    Generate a tensor train network
    """
    G = nx.DiGraph()
    G.add_nodes_from(range(f))

    # leaf edges
    add_leaves(G, f)

    # normal edges
    for i in range(f - 1):
        G.add_edge(i, i + 1)

    # reverse edges
    for i in range(f - 1, 0, -1):
        G.add_edge(i, i - 1)

    if isinstance(primitive_grid, int):
        Ns = [primitive_grid] * f
    else:
        # Ns = N
        Ns = []
        for grid in primitive_grid:
            Ns.append(len(grid))
            

    # add ranks
    for edge in sweep(G):
        if not is_leaf(edge, G):
            pre = pre_edges(G, edge, True)
            rmax = np.prod(collect(G, pre, 'r'))
            G.edges[edge]['r'] = min(r, rmax)
        else:
            coord = G.edges[edge]['coordinate']
            G.edges[edge]['r'] = Ns[coord]

    for edge in sweep(G):
        if not is_leaf(edge, G):
            r = G.edges[edge]['r']
            other = G.edges[flip(edge)]['r']
            G.edges[edge]['r'] = min(r, other)

    return G

def tensor_train_operator_graph(f, r = 2, N = 8):
    """
    Generate a tensor train network
    """
    G = nx.DiGraph()
    G.add_nodes_from(range(f))

    # leaf edges
    for i in range(0, f):
        node = i
        x = -2*i - 1
        y = -2*i - 2
        G.add_edge(node, x)
        G.add_edge(x, node)
        G.add_edge(node, y)
        G.add_edge(y, node)


    # normal edges
    for i in range(f - 1):
        G.add_edge(i, i + 1)

    # reverse edges
    for i in range(f - 1, 0, -1):
        G.add_edge(i, i - 1)

    # add ranks
    for edge in G.edges():
        if not is_leaf(edge, G):
            G.edges[edge]['r'] = r
        else:
            G.edges[edge]['r'] = N

    # add random edge entries
    coord = 0
    for edge in G.edges():
        if is_leaf(edge, G) and up_edge(edge, G):
            G.edges[edge]['coordinate'] = coord
            coord += 1
    return G

def _combine_nodes(nodes, id, edges):
    """
    Helper function for balanced_tree
    Combines nodes from a vector into new nodes.
    nodes: list of nodes
    id: new node idx
    nodes: {1, 2, 3, 4} -> {5, 6}
    n = 5 -> 7
    and add edges {(1, 5), (2, 5), (3, 6), (4, 6)} to G
    """
    for i in range(len(nodes) - 2, -1, -2):
        l = nodes[i]
        r = nodes[i+1]
        nodes.pop(i)
        nodes.pop(i)
        nodes.append(id)
        edges.append((r, id))
        edges.append((l, id))
        id += 1
    return nodes, id, edges

def build_tree(f):
    """
    create edges for a (close-to) balanced tree with f leaves
    """
    nodes = list(range(f))
    # special case for odd number of leaves
    if f % 2 == 1:
        nodes[0] = -1 
    id = f
    edges = []
    while len(nodes) > 1:
        nodes, id, edges = _combine_nodes(nodes, id, edges)
    return edges

def balanced_tree(f, r = 2, N = 8):
    """
    Generate a close-to balanced tree tensor network
    Node: odd f is implemented manually:
          avoids adding unnecessary node by added code.
          See # odd-leaf tag
    """
    G = nx.DiGraph()

    start = f % 2 # special case for odd number of leaves
    for i in range(start, f):
        edge = (-i - 1, i)
        G.add_edge(edge[0], edge[1])
        G.add_edge(edge[1], edge[0])

    edges = build_tree(f)
    for edge in edges:
        G.add_edge(edge[0], edge[1])

    for edge in reversed(edges):
#        if edge[0] < 0:
#            continue # special case for odd number of leaves
        G.add_edge(edge[1], edge[0])
    
    # add ranks
    for edge in sweep(G):
        if not is_leaf(edge, G):
            pre = pre_edges(G, edge, True)
            rmax = np.prod(collect(G, pre, 'r'))
            G.edges[edge]['r'] = min(r, rmax)
        else:
            G.edges[edge]['r'] = N

    for edge in sweep(G):
        if not is_leaf(edge, G):
            r = G.edges[edge]['r']
            other = G.edges[flip(edge)]['r']
            G.edges[edge]['r'] = min(r, other)

    return G 



import ase
from ase.io import write
import tempfile
import os
import re
import subprocess
import numpy as np
import torch


_xtb_method = {
    "gff": {"method_flags": ("--gff")},
    "gfn1": {"method_flags": ("--gfn", "1")},
    "gfn2": {"method_flags": ("--gfn", "2")},
}

def read_gradient(filename):
    with open(filename) as f:
        lines = [line.strip() for line in f if line.strip() and not line.lstrip().startswith('#')]
    natoms = int(lines[0])
    grad_vals = [float(x) for x in lines[2 : 2 + natoms * 3]]
    return np.array(grad_vals).reshape(natoms, 3)


def xtb(molecule: ase.Atoms, method: str,
        charge: int, n_threads: int, solvent: str = None,
        gradient: bool = False) -> tuple:
    """
    Run xTB (energy or gradient)

    Parameters
    ----------
    molecule : ase.Atoms
        Computes gradient for this Molecule

    Returns
    -------
    energy : float
        Energy in Hartree
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        # Write temporary file
        xyz_file = os.path.join(temp_dir, "tmp_input.xyz")
        write(xyz_file, molecule)
        flags = ["xtb", xyz_file]
        flags += [_xtb_method[method.lower()]["method_flags"]]
        flags += ["--charge", str(charge)]
        flags += ["--parallel", str(n_threads)]
        flags += ["--alpb", solvent] if solvent is not None else []
        if gradient:
            flags += ["--grad"]
        xtb_command = flags

        try:
            result = subprocess.run(
                xtb_command, check=True, capture_output=True, text=True, cwd=temp_dir
            )
        except subprocess.CalledProcessError as e:
            print("Error running xtb command:", e.stderr)
            raise e

        match = re.search(r"TOTAL ENERGY\s+([-+]?\d*\.\d+)\s+Eh", result.stdout)

        if match:
            energy = float(match.group(1))  # Extract and convert the energy to a float
        else:
            energy = None

        if gradient:
            dx = read_gradient(os.path.join(temp_dir, "tmp_input.engrad"))
            return energy, dx
        else:
            return energy, molecule.positions
        

class differentiable_xtb(torch.autograd.Function):
    """
    Autograd implementation of xTB
    """
    @staticmethod
    def forward(ctx, x, mol, method, charge, parallel, solvent = None):
        mol.positions = x.detach().numpy()
        e, _ = xtb(mol, method, charge, parallel, solvent, gradient = False)
        ctx.args = (mol, method, charge, parallel, solvent)
        ctx.save_for_backward(x)
        return torch.tensor([e], dtype=x.dtype, device=x.device)

    @staticmethod
    def backward(ctx, grad_output):
        mol, method, charge, parallel, solvent = ctx.args
        (x,) = ctx.saved_tensors
        mol.positions = x.detach().numpy()
        _, dx = xtb(mol, method, charge, parallel, solvent, gradient = True)
        grad_x = grad_output * torch.tensor(dx, dtype=grad_output.dtype, device=grad_output.device)
        return grad_x, None, None, None, None, None


def xtb_lbfgs(x, mol, method = 'gff', charge=0, parallel=1, solvent=[], max_iter = 100, lr = 0.1):
    """
    run L-BFGS on xTB using autograd

    x: torch tensor cartesian coordinates
    mol: ase.atoms molecule
    method (gff|gfn1|gfn2): xTB method
    max_iter: maximal number of iterations
    lr: initial step length
    """
    optimizer = torch.optim.LBFGS([x], max_iter=max_iter, lr=lr)
    
    def closure():
        optimizer.zero_grad()
        energy = differentiable_xtb.apply(x, mol, method, charge, parallel)
        energy.backward()
        return energy
    
    optimizer.step(closure)




import torch
from pyscf import gto



class ERI(torch.autograd.Function):
    @staticmethod
    def forward(ctx, coords, atom_charges, basis):
        coords_np = coords.detach().cpu().numpy()
        atom_charges = atom_charges.detach().cpu().numpy().astype(int)

        mol = gto.Mole()
        mol.atom = [[int(z), tuple(r)] for z, r in zip(atom_charges, coords_np)]
        mol.basis = basis
        mol.build()

        eri = mol.intor('int2e')
        d_eri = mol.intor('int2e_ip1')  # gradient of (ij|kl) w.r.t. nuclear coords

        ctx.save_for_backward(torch.tensor(d_eri))
        return torch.tensor(eri)

    # @staticmethod
    # def backward(ctx, grad_output):
    #     (d_eri,) = ctx.saved_tensors
    #     # grad_output: shape (nao, nao, nao, nao)
    #     # d_eri: shape (natm, 3, nao, nao, nao, nao)
    #     grad_coords = torch.einsum('aijkl,nijkl->na', grad_output, d_eri)
    #     return grad_coords, None, None

    @staticmethod
    def backward(ctx, grad_output):
        (d_eri,) = ctx.saved_tensors
        print(grad_output)
        print(d_eri.shape)
        
        grad_coords = torch.einsum('ijkl,najkl->an', grad_output, d_eri)
        # return None, None, None
        return grad_coords, None, None





import numpy as np

def givens_rotation(B, c, s, i, j):
    """
    Perform a Givens-Rotation: R*A*R^H with R= {{c, s}, {-s*, c}}
    """
    copy0 = (c * B[:, :, i] + s * B[:, :, j]).copy()
    copy1 = (c * B[:, :, j] - np.conj(s) * B[:, :, i]).copy()
    B[:, :, i] = copy0
    B[:, :, j] = copy1
    copy0 = B[:, i, :].copy()
    copy1 = B[:, j, :].copy()
    B[:, i, :] = c * copy0 + np.conj(s) * copy1
    B[:, j, :] = c * copy1 - s * copy0

def build_G_matrix_real(A, i, j):
    """
    Build the G-Matrix
    """
    M = A.shape[0]
    G = np.zeros((2, 2), dtype=np.float64)
    h = np.zeros((M, 2), dtype=np.complex128)

    # Build h-Vector for this Matrix
    h[:, 0] = A[:, i, i] - A[:, j, j]
    h[:, 1] = A[:, i, j] + A[:, j, i]

    G = np.real(np.einsum('mi,mj->ij', np.conj(h), h))
    return G

def calculate_angles_real(A, i, j):
    """
    Calculate the angles c and s for the Givens-Rotation
    """
    # Build the G-Matrix
    G = build_G_matrix_real(A, i, j)

    # Diagonalize it (phase convention is important here (x_>0)!)
    _, trafo = np.linalg.eigh(G)

    # Calculate the angles from the
    x = trafo[0, 1]
    y = trafo[1, 1]

    r = np.sqrt(x * x + y * y)
    t = np.sqrt(2 * r * (x + r))
    if r == 0 or t == 0:
        return 1., 0.

    c = np.sqrt((x + r) / (2*r))
    s = y  / t

    norm = c * c + s * s
    c /= np.sqrt(norm)
    s /= np.sqrt(norm)
    return c, s

def build_G_matrix(A, i, j):
    """
    Build the G-Matrix
    """
    M = A.shape[0]
    G = np.zeros((3, 3), dtype=np.float64)
    h = np.zeros((M, 3), dtype=np.complex128)

    # Build h-Vector for this Matrix
    h[:, 0] = A[:, i, i] - A[:, j, j]
    h[:, 1] = A[:, i, j] + A[:, j, i]
    h[:, 2] = 1j * (A[:, j, i] - A[:, i, j])

    G = np.real(np.einsum('mi,mj->ij', np.conj(h), h))
    return G

def calculate_angles(A, i, j):
    """
    Calculate the angles c and s for the Givens-Rotation
    """
    # Build the G-Matrix
    G = build_G_matrix(A, i, j)

    # Diagonalize it (phase convention is important here (x_>0)!)
    _, trafo = np.linalg.eigh(G)

    # Calculate the angles from the
    x = trafo[0, 2]
    y = trafo[1, 2]
    z = trafo[2, 2]

    r = np.sqrt(x * x + y * y + z * z)
    t = np.sqrt(2 * r * (x + r))
    if np.abs(r) < 1e-16:
        C = np.array([A[0, i, i], A[0, i, j], A[0, j, i], A[0, j, j]]).reshape((2, 2))
        h = np.zeros((A.shape[0], 3), dtype=np.complex128)
        h[:, 0] = A[:, i, i] - A[:, j, j]
        h[:, 1] = A[:, i, j] + A[:, j, i]
        h[:, 2] = 1j * (A[:, j, i] - A[:, i, j])
        print(i, j, A.dtype)
        print('A[] = ', C)
        print('h = ', h)
        print(G)
        print(x, y, r, t)
        return 1., 0.

    #c = (x + r) / t
    if r == 0 or t == 0:
        return 1., 0.
    c = np.sqrt((x + r) / (2*r))
    s = (y - 1j * z) / t
    if np.isrealobj(A):
        s = y / t
    else:
        s = (y - 1j * z) / t
    norm = c * c + s * s
    c /= np.sqrt(norm)
    s /= np.sqrt(norm)
    return c, s

def givens_trafo_rotation(U, c, s, i, j):
    """
    Perform a Givens-Rotation on the Transformation-Matrix
    """
    Ui = U[:, i].copy()
    Uj = U[:, j].copy()
    copy0 = (c * Ui + s * Uj)
    copy1 = (c * Uj - np.conj(s) * Ui)
    U[:, i] = copy0.copy()
    U[:, j] = copy1.copy()
    return U

def jacobi_rotations(A, U):
    """
    Perform Jacobi-Rotations to simultaneously diagonalize a set of matrices
    """
    # Angles for Givens-Rotation
    c, s = 0, 0

    acc = 0. # accumulated off-angles
    # Swipe over the matrix-dimensions and perform jacobi-rotations
    for i in range(A.shape[1]):
        for j in range(i + 1, A.shape[2]):
            # Calculate Angles c and s for the elements i and j
            c, s = 1., 0.
            if np.isrealobj(A):
                print('real')
                c, s = calculate_angles_real(A, i, j)
            else:
                print('complex')
                c, s = calculate_angles(A, i, j)
            acc += np.abs(s) # add rotations of off-diagonal elements

            # Perform the Givens-Rotation with angles c and s
            givens_rotation(A, c, s, i, j)

            # Rotate the Transformation-Matrix
            givens_trafo_rotation(U, c, s, i, j)
    return acc

def off_diagonal_measure(A):
    A = A.copy()
    result = 0.
    for i in range(A.shape[0]):
        np.fill_diagonal(A[i], 0.)
        result += np.linalg.norm(A[i])
    return result

def simultaneous_diagonalization(A, n_iter = 100, eps = 1e-12, verbose = False, copy = True):
    """
    A: np.ndarray with shape (k, m, m)
    k: number of matrices to diagonalize simultaneously
    eps: convergence criterion
    n_iter: maximum number of iterations
    verbose: print progress of the algorithm
    copy: if True, the input array is copied
    """
    if len(A.shape) != 3:
        raise ValueError('A must be a 3D tensor.')
    if copy:
        A = A.copy()

    U = np.eye(A.shape[1], dtype=A.dtype)
    for it in range(n_iter):
        acc = jacobi_rotations(A, U)
        delta = off_diagonal_measure(A)
        if verbose:
            print('Iteration: {} Delta: {} Acc: {}'.format(it, delta, acc))
        if delta < eps:
            break

    return A, U


import numpy as np


def V(x):
    N = x.shape[0]
    v = 0.
    for i in range(N):
        v += 0.5*(x[i])**2 + 0.1*x[i]*x[(i+1) % N]
    return v


def filterV(x):
    return np.exp(-V(x))


def pot(x):
    v = 0.
    for i in range(0, x.shape[0]):
        v += np.sin(np.pi*x[i]**2)/(x[i]**2+1e-7)
    for i in range(0, x.shape[0]):
        v+=np.sin(x[i]*x[i-1])
    return v - 15.


def rastigrin(x):
    y = 5.12 * (x - 0.5) * 2
    A = 10
    n = len(x)
    v = n * A
    for i in range(n):
        v += (y[i]**2 - A * np.cos(np.pi * 2 * y[i]))
    return v


def double_minimum(x):
    x = np.asarray(x)
    return np.sum((x - 0.1)**2 * (x - 0.8)**2) + 1


def double_minimum_nd(x):
    x = np.asarray(x)
    well1 = np.sum((x - 0.1)**2) + 0.1
    well2 = np.sum((x - 0.8)**2)
    return np.minimum(well1, well2) + 0.1 * well1


def double_minimum_exp(x):
    x = np.asarray(x)
    well1 = (np.exp(-np.sum(3*(x - 0.1)**2))-1)
    well2 = (np.exp(-np.sum(5*(x - 0.8)**2)))*0.8
    return -(well1 + well2)


def rosenbrock_function(x):
    return np.sum(100. * (x[1:] - x[:-1]**2)**2 + (1 - x[:-1])**2)




import torch


"""
Torch version of BAT coordinates for automated differentiation
Translated from ASE implementation with slight adjustments
"""

def get_dihedral(X, a0, a1, a2, a3):
    """
    Compute dihedral for vertices a0, a1, a2, a3 in X
    
    X: cartesian coordinates
    a0, a1, a2, a3: indices of atoms
    """
    b0 = X[a1] - X[a0]
    b1 = X[a2] - X[a1]
    b2 = X[a3] - X[a2]

    b1_norm = b1 / torch.norm(b1)

    # Project b0 and b2 onto plane perpendicular to b1
    v = -b0 - torch.dot(-b0, b1_norm) * b1_norm
    w = b2 - torch.dot(-b2, b1_norm) * b1_norm

    if torch.allclose(v, torch.zeros_like(v)) or torch.allclose(w, torch.zeros_like(w)):
        raise ZeroDivisionError("Undefined dihedral for planar inner angle")

    x = torch.dot(v, w)
    y = torch.linalg.cross(b1_norm, v).dot(w)
    angle = torch.atan2(y, x)
    return torch.rad2deg(angle % (2 * torch.pi))


def masked_rotate(X, center, axis, angle_rad, mask):
    """
    Rotate mask-atoms in X atound axis. Center has to be point on axis
    center: (3,)
    axis: (3,), not necessarily normalized
    angle_rad: scalar
    mask: (N,) boolean
    X: (N, 3)
    """

    axis = axis / torch.norm(axis)

    # Select subgroup
    subgroup = X[mask] - center  # translate to origin

    # Rodrigues' rotation formula
    cos = torch.cos(angle_rad)
    sin = torch.sin(angle_rad)
    dot = (subgroup * axis).sum(dim=1, keepdim=True)
    rotated = (
        subgroup * cos
        + torch.cross(axis.expand_as(subgroup), subgroup, dim=1) * sin
        + axis * dot * (1 - cos)
    )

    # Translate back and write to X
    X[mask] = rotated + center


def set_dihedral(X, a0, a1, a2, a3, angle, mask=None, indices=None):
    """
    Set dihedral in X
    
    X: atom positions
    a0, a1, a2, a3: indices o atoms
    angle: taget value for dihedral
    mask: target atom indices that will move
    """
    # angle = angle_deg * torch.pi / 180

    if mask is None and indices is None:
        mask = torch.zeros(len(X), dtype=torch.bool, device=X.device)
        mask[a3] = True
    elif indices is not None:
        mask = torch.zeros(len(X), dtype=torch.bool, device=X.device)
        mask[indices] = True
    else:
        mask = torch.as_tensor(mask, dtype=torch.bool, device=X.device)

    current = get_dihedral(X, a0, a1, a2, a3)
    diff = torch.deg2rad(angle - current)
    axis = X[a2] - X[a1]
    center = X[a2]
    masked_rotate(X, center, axis, diff, mask)


def moment_of_inertia(coords: torch.Tensor) -> torch.Tensor:
    coords = coords - coords.mean(dim=0, keepdim=True)
    x, y, z = coords[:, 0], coords[:, 1], coords[:, 2]
    Ixx = (y**2 + z**2).sum()
    Iyy = (x**2 + z**2).sum()
    Izz = (x**2 + y**2).sum()
    Ixy = -(x * y).sum()
    Ixz = -(x * z).sum()
    Iyz = -(y * z).sum()
    return torch.stack(
        [
            torch.stack([Ixx, Ixy, Ixz]),
            torch.stack([Ixy, Iyy, Iyz]),
            torch.stack([Ixz, Iyz, Izz]),
        ]
    )



from qutree.ttn.tensor_network import *
from qutree.ttn.grid import *

class Logger:
    def __init__(self):
        self.df = pd.DataFrame()
    
    def __call__(self, dic):
        self.df = pd.concat([self.df, pd.DataFrame(dic, index=[0])], ignore_index=True)

    def __str__(self):
        return f"Optimal value:\n{self.df.loc[self.df['f'].idxmin()]}"


def nparray_to_tuple(arr, precision = 8):
    return tuple(np.round(arr, precision))
    
class Objective:
    def __init__(self, Err, transformer = lambda x: x):
        self.Err = Err
        self.logger = Logger()
        self.cache = {}
        self.cache_hits = 0
        self.function_calls = 0
        self.transformer = transformer

    def __call__(self, x, **kwargs):
        # cache check
        key = nparray_to_tuple(x)
        if key in self.cache:
            self.cache_hits += 1
            return self.transformer(self.cache[key])

        f = self.Err(x)
        feff = self.transformer(f)

        # add to cache
        self.cache[key] = f
        self.function_calls += 1

        # logging
        if kwargs:
            xs = {f'x{i+1}': key[i] for i in range(len(x))}
            self.logger({**xs, 'f': f, **kwargs})
        return feff

    def __str__(self):
        nfunction = self.function_calls
        ncache = self.cache_hits
        output = str(self.logger) + f"""\n
Number of objective function calls: {nfunction}
Number of cached function accesses: {ncache}
Total number calls: {nfunction+ncache}"""
        return output

def ttnopt_step(G, O, sweep_id):
    G = G.copy()
    for edge in star_sweep(G, exclude_leafs=True):

        edges = pre_edges(G, edge)
        grids = collect(G, edges, 'grid')
        grid = cartesian_product(grids).permute()
        ranks = collect(G, edges, 'r')
        kwargs = {'sweep': sweep_id, 'node': edge[0]}
        A = quTensor(grid.evaluate(O, **kwargs).reshape(ranks), edges)

        next, cross_inv = maxvol_grids(A, G, edge)

        # save results
        G.edges[edge]['grid'] = next
        G.edges[edge]['A'] = cross_inv
        G.nodes[edge[0]]['grid'] = grid
        G.nodes[edge[0]]['A'] = A
    return G


def ttnopt(G: nx.DiGraph, O: Objective, nsweep: int = 6,
           primitive_grid: list[np.array] = None,
           start_grid = None):
    """
    Run tensor network optimization on objective function using provided graph and grids.
    
    Arguments:
    G (nx.DiGraph): tensor network (graph)
    O (Objective): objective function
    primitive grids (list[np.array[int]]): list of numpy arrays of 
    """
    if primitive_grid:
        G = tn_grid(G, primitive_grid, start_grid=start_grid)

    for sw in range(nsweep):
        G = ttnopt_step(G, O, sw)

    return G

def tn_CUR(G, O):
    # Fix the tensors at the nodes to get proper cross approximation of O
    # Will re-evaluate points that have been evaluated already but who cares.
    # G: Tensor network
    # O: Objective function (should have cache)
    for node in G.nodes():
        if node < 0:
            continue
        edges = G.in_edges(node)
        grids = collect(G, edges, 'grid')
        grid = cartesian_product(grids).permute()
        ranks = collect(G, edges, 'r')
        A = quTensor(grid.evaluate(O).reshape(ranks), edges)
        G.nodes[node]['A'] = A

    for edge in sweep(G, include_leaves=False):
        edges = pre_edges(G, edge)
        grids = collect(G, edges, 'grid')
        grid = cartesian_product(grids).permute()
        ranks = collect(G, edges, 'r')
        A = quTensor(grid.evaluate(O).reshape(ranks), edges)
        G[edge[0]][edge[1]]['A'] = A

    for edge in sweep(G, include_leaves=False):
        edges = [edge, flip(edge)]
        grids = collect(G, edges, 'grid')
        grid = cartesian_product(grids).permute()
        ranks = collect(G, edges, 'r')
        cross = grid.evaluate(O).reshape(ranks)
        Ainv = quTensor(regularized_inverse(cross, 1e-12), edges)
        G[edge[0]][edge[1]]['A'] = Ainv
    return G
